from __future__ import annotations

import logging
from contextlib import AsyncExitStack
from functools import cached_property

from aio_pika import (
    Exchange as AioPikaExchange,
    RobustChannel,
    connect_robust,
)
from aio_pika.abc import AbstractRobustConnection

from icij_common.logging_utils import LogWithNameMixin
from icij_worker import ManagerEvent
from . import EventPublisher
from ..routing_strategy import Routing
from ..utils.amqp import AMQPMixin, RobustConnection


class AMQPPublisher(AMQPMixin, EventPublisher, LogWithNameMixin):
    def __init__(
        self,
        logger: Optional[logging.Logger] = None,
        *,
        broker_url: str,
        connection_timeout_s: float = 1.0,
        reconnection_wait_s: float = 5.0,
        is_qpid: bool = False,
        app_id: str | None = None,
        connection: Optional[AbstractRobustConnection] = None,
    ):
        super().__init__(
            broker_url,
            connection_timeout_s=connection_timeout_s,
            reconnection_wait_s=reconnection_wait_s,
            is_qpid=is_qpid,
        )
        if logger is None:
            logger = logging.getLogger(__name__)
        LogWithNameMixin.__init__(self, logger)
        self._app_id = app_id
        self._broker_url = broker_url
        self._connection_ = connection
        self._channel_: Optional[RobustChannel] = None
        self._manager_evt_x: Optional[AioPikaExchange] = None
        self._connection_timeout_s = connection_timeout_s
        self._reconnection_wait_s = reconnection_wait_s
        self._exit_stack = AsyncExitStack()
        self._declare_and_bind = True

    async def __aenter__(self) -> AMQPPublisher:
        self.info("starting publisher connection workflow...")
        await self._exit_stack.__aenter__()
        await self._connection_workflow()
        self.info("publisher connected !")
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    @cached_property
    def _routings(self) -> list[Routing]:
        return [self.manager_evt_routing()]

    async def _publish_event(self, event: ManagerEvent):
        await self._publish_message(
            event,
            exchange=self._manager_evt_x,
            routing_key=self.manager_evt_routing().routing_key,
            mandatory=False,
        )

    async def _connection_workflow(self):
        self.debug("creating connection...")
        if self._connection_ is None:
            self._connection_ = await connect_robust(
                self._broker_url,
                timeout=self._connection_timeout_s,
                reconnect_interval=self._reconnection_wait_s,
                connection_class=RobustConnection,
            )
            await self._exit_stack.enter_async_context(self._connection)
        self.debug("creating channel...")
        self._channel_ = await self._connection.channel(
            publisher_confirms=self._publisher_confirms,
            on_return_raises=False,
        )
        await self._exit_stack.enter_async_context(self._channel)
        await self._channel_.set_qos(1, global_=False)
        if self._declare_and_bind:
            await self._create_routing(self.manager_evt_routing())
        self._manager_evt_x = await self._channel.get_exchange(
            self.manager_evt_routing().exchange.name, ensure=True
        )
        self.info("channel opened !")



import logging
from collections.abc import Iterable
from functools import partial

from fastapi import FastAPI, routing
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException as StarletteHTTPException

from icij_common.fastapi_utils import (
    http_exception_handler,
    internal_exception_handler,
    request_validation_error_handler,
)
from icij_worker.http_.constants import OTHER_TAG, TASKS_TAG
from icij_worker.http_.config import HttpServiceConfig
from icij_worker.http_.dependencies import BASE_DEPENDENCIES, run_http_service_deps
from icij_worker.http_.main import main_router
from icij_worker.http_.tasks import tasks_router
from icij_worker.typing_ import Dependency

RouterTag = str

INTERNAL_SERVER_ERROR = "Internal Server Error"
_REQUEST_VALIDATION_ERROR = "Request Validation Error"

logger = logging.getLogger(__name__)


def _make_open_api_tags(tags: Iterable[str]) -> list[dict]:
    return [{"name": t} for t in tags]


def create_service(
    config: HttpServiceConfig,
    routers: list[tuple[RouterTag, routing.APIRouter]] = None,
    dependencies: list[Dependency] | None = None,
) -> FastAPI:
    if routers is None:
        routers, tags = [], []
    else:
        tags, routers = zip(*routers)
        tags = list(tags)
        routers = list(routers)
    routers.append(main_router())
    routers.append(tasks_router())
    tags.extend([TASKS_TAG, OTHER_TAG])
    deps = BASE_DEPENDENCIES
    if dependencies is not None:
        deps += dependencies
    lifespan = partial(run_http_service_deps, config=config, dependencies=deps)
    app = FastAPI(
        title=config.app_title,
        openapi_tags=_make_open_api_tags(tags),
        lifespan=lifespan,
    )
    app.add_exception_handler(RequestValidationError, request_validation_error_handler)
    app.add_exception_handler(StarletteHTTPException, http_exception_handler)
    app.add_exception_handler(Exception, internal_exception_handler)
    for router in routers:
        app.include_router(router)
    return app



# pylint: disable=redefined-outer-name
# Test utils meant to be imported from clients libs to test their implem of workers
from __future__ import annotations

import asyncio
import logging
import signal
import uuid
from abc import ABC
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, ClassVar, TypeVar, cast

from pydantic import BaseModel, Field

import icij_worker
from icij_common.pydantic_utils import icij_config, ignore_extra_config
from icij_worker import (
    AsyncApp,
    AsyncBackend,
    FSKeyValueStorage,
    ManagerEvent,
    RoutingStrategy,
    Task,
    TaskState,
    Worker,
    WorkerConfig,
)
from icij_worker.app import AsyncAppConfig, TaskGroup
from icij_worker.event_publisher import EventPublisher
from icij_worker.exceptions import (
    MessageDeserializationError,
    TaskQueueIsFull,
    UnknownTask,
)
from icij_worker.objects import (
    CancelEvent,
    Message,
    ShutdownEvent,
    WorkerEvent,
)
from icij_worker.task_manager import TaskManager, TaskManagerConfig
from icij_worker.typing_ import RateProgress
from icij_worker.utils.dependencies import DependencyInjectionError
from icij_worker.utils.logging_ import LogWithWorkerIDMixin

logger = logging.getLogger(__name__)

_has_pytest = False  # necessary because of the pytest decorators which requires pytest
# to be defined
try:
    import pytest

    _has_pytest = True
except ImportError:
    pass

if _has_pytest:
    # TODO: make this one a MockStorage
    class DBMixin(FSKeyValueStorage, ABC):
        _locks_db_name = "locks"
        _manager_events_db_name = "manager_events"
        _worker_events_db_name = "worker_events"

        _routing_strategy: RoutingStrategy

        def __init__(self, db_path: Path) -> None:
            super().__init__(db_path)
            self._task_meta: dict[str, tuple[str, str]] = dict()

        def _make_group_dbs(self) -> dict:
            dbs = dict()
            for k in [
                self._tasks_db_name,
                self._results_db_name,
                self._errors_db_name,
                self._locks_db_name,
                self._manager_events_db_name,
                self._worker_events_db_name,
            ]:
                dbs[k] = self._make_db(self._db_path, name=k)
            return dbs

        @property
        def db_path(self) -> Path:
            return self._db_path

        @staticmethod
        def _order_events(events: list[dict]) -> float:
            events = [
                datetime.fromisoformat(evt["createdAt"]).timestamp() for evt in events
            ]
            return -min(events)

        async def _consume_(
            self,
            db_name: str,
            sleep_interval: float,
            factory: Callable[[Any], R],
            group: str | None = None,
            select: Callable[[R], bool] | None = None,
            order: Callable[[R], Any] | None = None,
        ) -> R:
            while "i'm waiting until I find something interesting":
                selected = dict(self._dbs[db_name].items())
                if db_name == self._tasks_db_name:
                    selected = {
                        k: v
                        for k, v in selected.items()
                        if k not in self._dbs[self._locks_db_name]
                    }
                if group is not None:
                    selected = [
                        (k, t) for k, t in selected.items() if t.get("group") == group
                    ]
                    for k, t in selected:
                        t.pop("group", None)
                else:
                    selected = selected.items()
                selected = [(k, factory(t)) for k, t in selected]
                if select is not None:
                    selected = [(k, t) for k, t in selected if select(t)]
                if selected:
                    if order is not None:
                        k, t = min(selected, key=lambda x: order(x[1]))
                    else:
                        k, t = selected[0]
                    return t
                await asyncio.sleep(sleep_interval)

    class MockAppConfig(BaseModel, LogWithWorkerIDMixin):
        model_config = icij_config()

        # Just provide logging stuff to be able to see nice logs while doing TDD
        log_level: str = "DEBUG"
        loggers: list[str] = [icij_worker.__name__]

    _MOCKED_CONFIG: MockAppConfig | None = None

    async def mock_async_config_enter(**_):
        global _MOCKED_CONFIG
        _MOCKED_CONFIG = MockAppConfig()
        logger.info(
            "Loading mocked configuration %s", _MOCKED_CONFIG.model_dump_json(indent=2)
        )

    def lifespan_config() -> MockAppConfig:
        if _MOCKED_CONFIG is None:
            raise DependencyInjectionError("config")
        return _MOCKED_CONFIG

    def loggers_enter(worker_id: str, **_):
        config = lifespan_config()
        config.setup_loggers(worker_id=worker_id)
        logger.info("worker loggers ready to log 💬")

    mocked_app_deps = [
        ("configuration loading", mock_async_config_enter, None),
        ("loggers setup", loggers_enter, None),
    ]

    APP = AsyncApp(name="test-app", dependencies=mocked_app_deps)

    async def _hello_world(greeted: str, progress: RateProgress | None = None) -> str:
        if progress is not None:
            await progress(0.1)
        greeting = f"Hello {greeted} !"
        if progress is not None:
            await progress(0.99)
        return greeting

    @APP.task
    async def hello_world(greeted: str, progress: RateProgress | None = None) -> str:
        return await _hello_world(greeted, progress)

    @APP.task(group="hello")
    async def grouped_hello_world(
        greeted: str, progress: RateProgress | None = None
    ) -> str:
        return await _hello_world(greeted, progress)

    @APP.task
    def hello_world_sync(greeted: str) -> str:
        greeting = f"Hello {greeted} !"
        return greeting

    async def _sleep_for(
        duration: float, s: float = 0.01, progress: RateProgress | None = None
    ):
        start = datetime.now()
        elapsed = 0
        while elapsed < duration:
            elapsed = (datetime.now() - start).total_seconds()
            await asyncio.sleep(s)
            if progress is not None:
                p = min(elapsed / duration, 1.0)
                await progress(p)

    @APP.task(max_retries=1)
    async def sleep_for(
        duration: float, s: float = 0.01, progress: RateProgress | None = None
    ):
        await _sleep_for(duration, s, progress)

    short_timeout = 1
    short_tasks_group = TaskGroup(name="short", timeout_s=short_timeout)

    @APP.task(max_retries=3, group=short_tasks_group)
    async def sleep_for_short(
        duration: float, s: float = 0.01, progress: RateProgress | None = None
    ):
        await _sleep_for(duration, s, progress)

    @APP.task(max_retries=666)
    async def often_retriable() -> str:
        pass

    @APP.task
    def case_test_task(snake_case_arg: str):
        return snake_case_arg

    @pytest.fixture(scope="session")
    def test_async_app() -> AsyncApp:
        return AsyncApp.load(f"{__name__}.APP")

    @pytest.fixture(scope="session")
    def app_config() -> AsyncAppConfig:
        return AsyncAppConfig()

    @pytest.fixture(scope="session")
    def late_ack_app_config() -> AsyncAppConfig:
        return AsyncAppConfig(late_ack=True)

    @pytest.fixture(scope="session")
    def test_async_app_late(late_ack_app_config: AsyncAppConfig) -> AsyncApp:
        return AsyncApp.load(f"{__name__}.APP", config=late_ack_app_config)

    class MockManagerConfig(TaskManagerConfig):
        backend: ClassVar[AsyncBackend] = Field(frozen=True, default=AsyncBackend.mock)
        db_path: Path
        event_refresh_interval_s: float = 0.1

    @TaskManager.register(AsyncBackend.mock)
    class MockManager(TaskManager, DBMixin):
        def __init__(
            self, app: AsyncApp, db_path: Path, event_refresh_interval_s: float = 0.1
        ):
            super().__init__(app)
            super(TaskManager, self).__init__(db_path)
            self._event_refresh_interval_s = event_refresh_interval_s

        @property
        def app(self) -> AsyncApp:
            return self._app

        async def _ensure_queue_size(self):
            tasks = await self.get_tasks(group=None, state=TaskState.QUEUED)
            n_queued = len(tasks)
            if (
                self.max_task_queue_size is not None
                and n_queued > self.max_task_queue_size
            ):
                raise TaskQueueIsFull(self.max_task_queue_size)

        async def _enqueue(self, task: Task):
            # pylint: disable=unused-argument
            await self._ensure_queue_size()

        async def _requeue(self, task: Task):
            await self._ensure_queue_size()
            group = await self.get_task_group(task.id)
            await self.save_task_(task, group)

        async def _consume(self) -> ManagerEvent:
            events = await self._consume_(
                self._manager_events_db_name,
                self._event_refresh_interval_s,
                list,
                select=lambda events: bool(  # pylint: disable=unnecessary-lambda
                    events
                ),
                order=self._order_events,
            )
            events = sorted(
                events,
                key=lambda e: datetime.fromisoformat(e["createdAt"]).timestamp(),
            )
            event = events[0]
            task_id = event["taskId"]
            key = self._key(task_id=task_id, obj_cls=ManagerEvent)
            manager_events_db = self._dbs[self._manager_events_db_name]
            manager_events_db[key] = events[1:]
            manager_events_db.commit(blocking=True)
            return cast(ManagerEvent, Message.model_validate(event))

        async def cancel(self, task_id: str, *, requeue: bool):
            cancel_event = CancelEvent(
                task_id=task_id, requeue=requeue, created_at=datetime.now(timezone.utc)
            )
            key = self._key(task_id=task_id, obj_cls=CancelEvent)
            worker_events = self._dbs[self._worker_events_db_name]
            events = worker_events.get(key, [])
            event_dict = cancel_event.model_dump(exclude_unset=True, by_alias=True)
            events.append(event_dict)
            worker_events[key] = events
            worker_events.commit(blocking=True)

        async def shutdown_workers(self):
            shutdown_event = ShutdownEvent(created_at=datetime.now(timezone.utc))
            key = self._key(task_id=f"shutdown-({uuid.uuid4()})", obj_cls=CancelEvent)
            worker_events = self._dbs[self._worker_events_db_name]
            events = worker_events.get(key, [])
            event_dict = shutdown_event.model_dump(by_alias=True, exclude_none=True)
            events.append(event_dict)
            worker_events[key] = events
            worker_events.commit(blocking=True)

        @classmethod
        def _from_config(cls, config: MockManagerConfig, **extras) -> MockManager:
            tm = cls(
                config.app,
                config.db_path,
                event_refresh_interval_s=config.event_refresh_interval_s,
            )
            return tm

        async def get_health(self) -> dict[str, bool]:
            return {"db": True}

    R = TypeVar("R")

    class MockEventPublisher(DBMixin, EventPublisher):
        _excluded_from_event_update = {"error"}

        def __init__(self, db_path: Path):
            super().__init__(db_path)
            self.published_events = []

        async def _publish_event(self, event: ManagerEvent):
            key = self._key(event.task_id, ManagerEvent)
            manager_events_db = self._dbs[self._manager_events_db_name]
            event_dict = event.model_dump(exclude_unset=True, by_alias=True)
            events = manager_events_db.get(key, [])
            events.append(event_dict)
            manager_events_db[key] = events
            manager_events_db.commit(blocking=True)
            self.published_events.append(event)

    def _default_loggers() -> list[str]:
        return [icij_worker.__name__]

    @WorkerConfig.register()
    class MockWorkerConfig(WorkerConfig):
        model_config = ignore_extra_config()

        type: AsyncBackend = Field(frozen=True, default=AsyncBackend.mock)

        db_path: Path
        log_level: str = "DEBUG"
        loggers: list[str] = Field(default_factory=_default_loggers)
        task_queue_poll_interval_s: float = 2.0

    @Worker.register(AsyncBackend.mock)
    class MockWorker(Worker, MockEventPublisher):
        def __init__(
            self,
            app: AsyncApp,
            worker_id: str | None = None,
            *,
            group: str | None,
            db_path: Path,
            poll_interval_s: float,
            **kwargs,
        ):
            MockEventPublisher.__init__(self, db_path)
            Worker.__init__(self, app, worker_id, group=group, **kwargs)
            self._poll_interval_s = poll_interval_s
            self._worker_id = worker_id
            self._logger_ = logging.getLogger(__name__)
            self.terminated_cancelled_event_loop = False
            self._exited = False

        @property
        def app(self) -> AsyncApp:
            return self._app

        @property
        def watch_events(self) -> asyncio.Task | None:
            return self._watch_events

        async def work_forever_async(self):
            await self._work_forever_async()

        @property
        def work_forever_task(self) -> asyncio.Task | None:
            return self._work_forever_task

        @property
        def work_once_task(self) -> asyncio.Task | None:
            return self._work_once_task

        @property
        def successful_exit(self) -> bool:
            return self._successful_exit

        @property
        def current(self) -> Task | None:
            return self._current

        @current.setter
        def current(self, value: Task | None):
            self._current = value

        @property
        def started_task_consumption(self) -> bool:
            return self._started_task_consumption_evt.is_set()

        async def signal_handler(self, signal_name: signal.Signals, *, graceful: bool):
            await self._signal_handler(signal_name, graceful=graceful)

        async def _aenter__(self):
            if not Path(self._db_path).exists():
                raise OSError(f"worker DB was not initialized ({self._db_path})")

        @classmethod
        def _from_config(cls, config: MockWorkerConfig, **extras) -> MockWorker:
            worker = cls(
                db_path=config.db_path,
                poll_interval_s=config.task_queue_poll_interval_s,
                **extras,
            )
            return worker

        async def _acknowledge(self, task: Task):
            key = self._key(task.id, Task)
            locks = self._dbs[self._locks_db_name]
            try:
                del locks[key]
            except KeyError as e:
                raise UnknownTask(task.id) from e
            locks.commit()

        async def _negatively_acknowledge(self, nacked: Task):
            key = self._key(nacked.id, Task)
            locks = self._dbs[self._locks_db_name]
            del locks[key]
            locks.commit()

        @staticmethod
        def _task_factory(obj: dict) -> Task:
            obj.pop("group", None)
            try:
                task = Task.model_validate(obj)
            except Exception as e:
                msg = f"invalid task object {obj}"
                raise MessageDeserializationError(msg) from e
            return task

        async def _consume(self) -> Task:
            task = await self._consume_(
                self._tasks_db_name,
                self._poll_interval_s,
                self._task_factory,
                group=self._group,
                select=lambda t: t.state is TaskState.QUEUED,
                order=lambda t: t.created_at,
            )
            locks = self._dbs[self._locks_db_name]
            key = self._key(task.id, Task)
            locks[key] = self._id
            locks.commit()
            return task

        async def _consume_worker_events(self) -> WorkerEvent:
            events = await self._consume_(
                self._worker_events_db_name,
                self._poll_interval_s,
                list,
                group=self._group,
                select=lambda evts: bool(evts),  # pylint: disable=unnecessary-lambda
                order=self._order_events,
            )
            events = sorted(
                events,
                key=lambda e: datetime.fromisoformat(e["createdAt"]).timestamp(),
            )
            event = events[0]
            try:
                event = cast(WorkerEvent, Message.model_validate(event))
            except Exception as e:
                msg = f"invalid event object {event}"
                raise MessageDeserializationError(msg) from e
            if isinstance(event, CancelEvent):
                worker_events = self._dbs[self._worker_events_db_name]
                key = self._key(event.task_id, WorkerEvent)
                worker_events[key] = events[1:]
                worker_events.commit(blocking=True)
            return event

        async def work_once(self):
            await self._work_once()

        async def publish_cancelled_event(self, requeue: bool):
            await self._publish_cancelled_event(requeue)



from __future__ import annotations

import asyncio
import inspect
import sys
from functools import partial, wraps
from typing import Any, Callable

from typer import Typer


class AsyncTyper(Typer):
    @staticmethod
    def maybe_run_async(decorator: Callable, func: Callable) -> Any:
        if inspect.iscoroutinefunction(func):

            @wraps(func)
            def runner(*args: Any, **kwargs: Any) -> Any:
                return asyncio.run(func(*args, **kwargs))

            decorator(runner)
        else:
            decorator(func)
        return func

    def callback(self, *args: Any, **kwargs: Any) -> Any:
        decorator = super().callback(*args, **kwargs)
        return partial(self.maybe_run_async, decorator)

    def command(self, *args: Any, **kwargs: Any) -> Any:
        decorator = super().command(*args, **kwargs)
        return partial(self.maybe_run_async, decorator)


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)  # noqa: F821



from __future__ import annotations

import functools
from typing import Callable

import neo4j
from neo4j.exceptions import ConstraintError, ResultNotSingleError
from pydantic import Field

from icij_worker import AsyncApp, AsyncBackend, Task, TaskState, Worker, WorkerConfig
from icij_worker.constants import (
    NEO4J_SHUTDOWN_EVENT_CREATED_AT,
    NEO4J_SHUTDOWN_EVENT_NODE,
    NEO4J_TASK_CANCELLED_BY_EVENT_REL,
    NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT,
    NEO4J_TASK_CANCEL_EVENT_EFFECTIVE,
    NEO4J_TASK_CANCEL_EVENT_NODE,
    NEO4J_TASK_GROUP,
    NEO4J_TASK_LOCK_NODE,
    NEO4J_TASK_LOCK_TASK_ID,
    NEO4J_TASK_LOCK_WORKER_ID,
    NEO4J_TASK_NODE,
    NEO4J_WORKER_ID,
    NEO4J_WORKER_NODE,
    NEO4J_WORKER_SHUTDOWN_BY_REL,
)
from icij_worker.event_publisher.neo4j_ import Neo4jEventPublisher
from icij_worker.exceptions import (
    MessageDeserializationError,
    TaskAlreadyReserved,
    UnknownTask,
)
from icij_worker.objects import CancelEvent, ShutdownEvent, WorkerEvent
from icij_worker.utils.neo4j_ import Neo4jConsumerMixin


@WorkerConfig.register()
class Neo4jWorkerConfig(WorkerConfig):
    type: AsyncBackend = Field(frozen=True, default=AsyncBackend.neo4j)

    cancelled_tasks_refresh_interval_s: float = 0.1
    poll_interval_s: float = 0.1
    neo4j_connection_timeout: float = 5.0
    neo4j_host: str = "127.0.0.1"
    neo4j_password: str | None = None
    neo4j_port: int = 7687
    neo4j_uri_scheme: str = "bolt"
    neo4j_user: str | None = None

    @property
    def neo4j_uri(self) -> str:
        return f"{self.neo4j_uri_scheme}://{self.neo4j_host}:{self.neo4j_port}"

    def to_neo4j_driver(self) -> neo4j.AsyncDriver:
        auth = None
        if self.neo4j_password:
            # TODO: add support for expiring and auto renew auth:
            #  https://neo4j.com/docs/api/python-driver/current/api.html
            #  #neo4j.auth_management.AuthManagers.expiration_based
            auth = neo4j.basic_auth(self.neo4j_user, self.neo4j_password)
        driver = neo4j.AsyncGraphDatabase.driver(
            self.neo4j_uri,
            connection_timeout=self.neo4j_connection_timeout,
            connection_acquisition_timeout=self.neo4j_connection_timeout,
            max_transaction_retry_time=self.neo4j_connection_timeout,
            auth=auth,
        )
        return driver


def _no_filter(group: str) -> bool:
    # pylint: disable=unused-argument
    return True


@Worker.register(AsyncBackend.neo4j)
class Neo4jWorker(Worker, Neo4jEventPublisher, Neo4jConsumerMixin):
    def __init__(
        self,
        app: AsyncApp,
        worker_id: str | None = None,
        *,
        group: str | None,
        driver: neo4j.AsyncDriver,
        poll_interval_s: float,
        **kwargs,
    ):
        super().__init__(app, worker_id, group=group, **kwargs)
        super(Worker, self).__init__(driver)
        if self._group is not None:
            db_filter = self._routing_strategy.db_filter_factory(self._group)
        else:
            db_filter = _no_filter
        self._db_filter: Callable[[str], bool] = db_filter
        self._poll_interval_s = poll_interval_s

    @classmethod
    def _from_config(cls, config: Neo4jWorkerConfig, **extras) -> Neo4jWorker:
        tasks_refresh_interval_s = config.cancelled_tasks_refresh_interval_s
        worker = cls(
            driver=config.to_neo4j_driver(),
            poll_interval_s=config.poll_interval_s,
            cancelled_tasks_refresh_interval_s=tasks_refresh_interval_s,
            **extras,
        )
        return worker

    async def _consume(self) -> Task:
        task = await self._consume_(
            functools.partial(_consume_task_tx, group=self._group, worker_id=self.id),
            refresh_interval_s=self._poll_interval_s,
            db_filter=self._db_filter,
        )
        try:
            task = Task.from_neo4j({"task": task})
        except Exception as e:
            msg = f"invalid task {task}"
            raise MessageDeserializationError(msg) from e
        return task

    async def _consume_worker_events(self) -> WorkerEvent:
        return await self._consume_(
            functools.partial(
                _consume_worker_events_tx, worker_id=self.id, group=self._group
            ),
            refresh_interval_s=self._poll_interval_s,
            db_filter=self._db_filter,
        )

    async def _acknowledge(self, task: Task):
        async with self._task_session(task.id) as sess:
            await sess.execute_write(
                _acknowledge_task_tx, task_id=task.id, worker_id=self.id
            )

    async def _negatively_acknowledge(self, nacked: Task):
        async with self._task_session(nacked.id) as sess:
            await sess.execute_write(
                _unlock_task_tx, task_id=nacked.id, worker_id=self.id
            )

    async def _aexit__(self, exc_type, exc_val, exc_tb):
        await self._driver.__aexit__(exc_type, exc_val, exc_tb)


async def _consume_task_tx(
    tx: neo4j.AsyncTransaction, *, worker_id: str, group: str | None
) -> Optional[neo4j.Record]:
    where_ns = ""
    if group is not None:
        where_ns = f"WHERE t.{NEO4J_TASK_GROUP} = $group"
    query = f"""MATCH (t:{NEO4J_TASK_NODE}:`{TaskState.QUEUED.value}`)
{where_ns}
WITH t
LIMIT 1
CALL apoc.create.setLabels(t, $labels) YIELD node AS task
WITH task
CREATE (lock:{NEO4J_TASK_LOCK_NODE} {{
    {NEO4J_TASK_LOCK_TASK_ID}: task.id,
    {NEO4J_TASK_LOCK_WORKER_ID}: $workerId 
}})
RETURN task"""
    labels = [NEO4J_TASK_NODE, TaskState.RUNNING.value]
    res = await tx.run(query, workerId=worker_id, labels=labels, group=group)
    try:
        task = await res.single(strict=True)
    except ResultNotSingleError:
        return None
    except ConstraintError as e:
        raise TaskAlreadyReserved() from e
    return task["task"]


async def _consume_worker_events_tx(
    tx: neo4j.AsyncTransaction, *, worker_id: str, group: str | None, **_
) -> Optional[WorkerEvent]:
    await _create_worker_node_tx(tx, worker_id)
    shutdown_event = await _consume_shutdown_events_tx(tx, worker_id)
    if shutdown_event is not None:
        return shutdown_event
    return await _consume_cancelled_events_tx(tx, group)


async def _create_worker_node_tx(tx: neo4j.AsyncTransaction, worker_id: str, **_):
    create_worker_node_query = f"""MERGE (
    worker: {NEO4J_WORKER_NODE} {{ {NEO4J_WORKER_ID}: $workerId }})
RETURN worker
"""
    await tx.run(create_worker_node_query, workerId=worker_id)


async def _consume_shutdown_events_tx(
    tx: neo4j.AsyncTransaction, worker_id: str, **_
) -> Optional[ShutdownEvent]:
    shutdown_event_query = f"""MATCH (worker: {NEO4J_WORKER_NODE} {{
    {NEO4J_WORKER_ID}: $workerId }})
WITH worker
MATCH (event:{NEO4J_SHUTDOWN_EVENT_NODE})
WHERE NOT (worker)-[:{NEO4J_WORKER_SHUTDOWN_BY_REL}]->(event)
WITH event, worker
ORDER BY event.{NEO4J_SHUTDOWN_EVENT_CREATED_AT} ASC
LIMIT 1
CREATE (worker)-[:{NEO4J_WORKER_SHUTDOWN_BY_REL}]->(event)
RETURN event
"""
    res = await tx.run(shutdown_event_query, workerId=worker_id)
    try:
        event = await res.single(strict=True)
    except ResultNotSingleError:
        return None
    try:
        event = ShutdownEvent.from_neo4j(event)
    except Exception as e:
        msg = f"invalid shutdown event {event}"
        raise MessageDeserializationError(msg) from e
    return event


async def _consume_cancelled_events_tx(
    tx: neo4j.AsyncTransaction, group: str | None, **_
):
    where_group = ""
    if group is not None:
        where_group = f"AND task.{NEO4J_TASK_GROUP} = $group"
    get_event_query = f"""MATCH (task:{NEO4J_TASK_NODE})-[
    :{NEO4J_TASK_CANCELLED_BY_EVENT_REL}
]->(event:{NEO4J_TASK_CANCEL_EVENT_NODE})
WHERE NOT event.{NEO4J_TASK_CANCEL_EVENT_EFFECTIVE}{where_group}
SET event.{NEO4J_TASK_CANCEL_EVENT_EFFECTIVE} = true
RETURN task, event
ORDER BY event.{NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT} ASC
LIMIT 1
"""
    res = await tx.run(get_event_query, group=group)
    try:
        event = await res.single(strict=True)
    except ResultNotSingleError:
        return None
    try:
        event = CancelEvent.from_neo4j(event)
    except Exception as e:
        msg = f"invalid cancel event {event}"
        raise MessageDeserializationError(msg) from e
    return event


async def _acknowledge_task_tx(
    tx: neo4j.AsyncTransaction, *, task_id: str, worker_id: str
):
    query = f"""MATCH (lock:{NEO4J_TASK_LOCK_NODE} {{
    {NEO4J_TASK_LOCK_TASK_ID}: $taskId 
}})
WHERE lock.{NEO4J_TASK_LOCK_WORKER_ID} = $workerId
DELETE lock
RETURN lock"""
    res = await tx.run(query, taskId=task_id, workerId=worker_id)
    try:
        await res.single(strict=True)
    except ResultNotSingleError as e:
        raise UnknownTask(task_id, worker_id) from e


async def _unlock_task_tx(tx: neo4j.AsyncTransaction, *, task_id: str, worker_id: str):
    query = f"""MATCH (lock:{NEO4J_TASK_LOCK_NODE} {{
     {NEO4J_TASK_LOCK_TASK_ID}: $taskId 
 }})
WHERE lock.{NEO4J_TASK_LOCK_WORKER_ID} = $workerId
DELETE lock
RETURN lock
"""
    res = await tx.run(query, taskId=task_id, workerId=worker_id)
    try:
        await res.single(strict=True)
    except ResultNotSingleError as e:
        raise UnknownTask(task_id, worker_id) from e



from fastapi import APIRouter
from starlette.responses import Response
from starlette.status import HTTP_200_OK, HTTP_503_SERVICE_UNAVAILABLE

from icij_worker.http_.constants import OTHER_TAG
from icij_worker.http_.dependencies import lifespan_task_manager


def main_router(app_version: str | None = None) -> APIRouter:
    router = APIRouter(tags=[OTHER_TAG])
    if app_version is not None:

        @router.get("/version")
        def version() -> Response:
            return Response(content=app_version, media_type="text/plain")

    @router.get("/health")
    async def health(response: Response) -> dict[str, bool]:
        task_manager = lifespan_task_manager()
        health = await task_manager.get_health()
        is_healthy = all(health.values())
        response.status_code = (
            HTTP_200_OK if is_healthy else HTTP_503_SERVICE_UNAVAILABLE
        )
        return health

    return router



from __future__ import annotations

import argparse
import logging
import sys
import traceback
from pathlib import Path

import uvicorn

import icij_worker
from icij_common.import_utils import import_variable
from icij_common.logging_utils import setup_loggers
from icij_worker.http_.config import HttpServiceConfig
from icij_worker.http_.service import create_service


class Formatter(argparse.ArgumentDefaultsHelpFormatter):
    def __init__(self, prog: str):
        super().__init__(prog, max_help_position=35, width=150)


def _start_app_(ns: argparse.Namespace) -> None:
    config_cls = None
    if ns.config_class is not None:
        config_cls = import_variable(ns.config_class)
    _start_app(config_cls=config_cls, config_path=ns.config_path)


def _start_app(
    config_cls: type[HttpServiceConfig] | None, config_path: str | None = None
) -> None:
    if config_cls is None:
        config_cls = HttpServiceConfig
    if config_path is not None:
        config = config_cls.model_validate_json(Path(config_path).read_text())
    else:
        config = config_cls()
    fast_api = create_service(config)
    log_level = logging.getLevelName(config.log_level)
    uvicorn.run(
        fast_api,
        host=config.host,
        port=config.port,
        workers=config.n_workers,
        log_level=log_level,
    )


def get_arg_parser() -> argparse.ArgumentParser:
    arg_parser = argparse.ArgumentParser(
        description="HTTP server start CLI", formatter_class=Formatter
    )
    arg_parser.add_argument("--config-class", type=str)
    arg_parser.add_argument("--config-path", type=str)
    arg_parser.set_defaults(func=_start_app_)
    return arg_parser


def main() -> None:
    # Setup loggers temporarily before loggers init using the app configuration
    setup_loggers(["__main__", icij_worker.__name__])
    logger = logging.getLogger(__name__)
    try:
        arg_parser = get_arg_parser()
        args = arg_parser.parse_args()

        if hasattr(args, "func"):
            args.func(args)
        else:
            arg_parser.print_help()
            sys.exit(1)
    except KeyboardInterrupt as e:
        logger.error("Application shutdown...")
        raise e
    except Exception as e:
        error_with_trace = "".join(traceback.format_exception(None, e, e.__traceback__))
        logger.error("Error occurred at application startup:\n%s", error_with_trace)
        raise e
